import pytest
from pytorch_quik.travel import QuikTraveler
from torch.utils import data
import torch.utils.data as td


def test_quik_traveler(args, gpus):
    myqt = QuikTraveler(args, 0)
    assert isinstance(myqt, QuikTraveler)
    assert isinstance(myqt.epochs, int)


def test_quik_data(gpus, sample_tensor, sample_labels, args):
    args.gpus = gpus
    if args.gpus == 0:
        pytest.skip()
    myqt = QuikTraveler(args, 0)
    mytds = td.TensorDataset(sample_tensor, sample_tensor, sample_labels)
    myqt.add_data(mytds)
    assert(isinstance(myqt.data.dataset, td.TensorDataset))
    assert(isinstance(myqt.data.data_loader, data.DataLoader))



import json
from pathlib import Path
from pytorch_quik import bert, io
from typing import Optional, List, OrderedDict, KeysView
from argparse import Namespace
import shlex
import subprocess
import requests
from urllib.parse import urlparse

EXTRA_FILES = [
    "config.json",
    "setup_config.json",
    "index_to_name.json",
    "tokenizer_config.json",
    "tokenizer.json",
    "special_tokens_map.json",
    "vocab.txt",
]


def save_setup_config(serve_path: str, labels: KeysView, args: Namespace):
    """Create a setup_config json for the huggingface torch serve handler

    Args:
        serve_path (str): The directory to store the sample.
        labels (KeysView): the dictionary keys of index_labels
        args (Namespace): the project argparse namespace.
    """
    serve_config = bert.get_bert_info(labels, args.bert_type)
    io.json_write(serve_path, "setup_config.json", serve_config)


def save_index_to_name(serve_path: str, indexed_labels: OrderedDict[int, str]):
    """Create the required index_to_name file for serving

    Args:
        serve_path (str): The directory to store the sample.
        indexed_labels (OrderedDict[str, int]): the target labels with indexes.
    """
    io.json_write(serve_path, "index_to_name.json", indexed_labels)


def inference_data(data: List[str]) -> str:
    """Format any data into an inference-ready json

    Args:
        data (List[str]): a list of inputs

    Returns:
        str: the final json-ready inference data.
    """
    dlist = [{"data": text} for text in data]
    data = f'{{"instances": {json.dumps(dlist)}}}'
    return data


def save_sample(serve_path):
    """A sample input to test the serving model

    Args:
        serve_path (str): The directory to store the sample.
    """
    sample = ["Great company with fast support"]
    sample = inference_data(sample)
    io.json_write(serve_path, "sample_text.json", sample)


def save_handler(serve_path, url: Optional[str] = None):
    """Download the handler file for serving.

    Args:
        serve_path ([type]): the torch serve directory
        url (str, optional): The url where the handler can be found.
        Defaults to None.
    """
    if url is None:
        url = urlparse("https://raw.githubusercontent.com")
        handler_loc = Path(
            "donchesworth/pytorch-quik/main/",
            "pytorch_quik",
            "handler",
            "transformer_handler_pq.py",
        )
        url = url._replace(path=str(handler_loc))
    filename = Path(serve_path).joinpath(handler_loc.name)
    r = requests.get(url.geturl(), allow_redirects=True)
    open(filename, "wb").write(r.content)


def mar_files(mar_path: Path, sfile: str, hfile: str) -> bool:
    """check of all mar files exist before attempting to
    run torch-model-archiver

    Args:
        mar_path (Path): the directory containing all the mar files.
        sfile (str): the serialized filename
        hfile (str): the handler filename

    Returns:
        bool: Whether all mar files were found.
    """
    all_files = [sfile, hfile]
    all_files.extend(EXTRA_FILES)
    missing = [
        file for file in all_files if not mar_path.joinpath(file).is_file()
    ]
    if len(missing) == 0:
        return True
    else:
        print("These files are missing:")
        print(missing)
        return False


def build_extra_files(
    args: Namespace,
    indexed_labels: OrderedDict[str, int],
    serve_path: Optional[Path] = None,
):
    """build all the boiler plate files to create
    an mar

    Args:
        args (Namespace): the project argparse namespace.
        indexed_labels (OrderedDict[str, int]): the target labels with indexes.
        serve_path (Optional[Path], optional): The diretory to contain
        all serve files. Defaults to None.
    """
    if serve_path is None:
        serve_path = io.id_str("state_dict", args).parent.joinpath("serve")
    serve_path.mkdir(parents=True, exist_ok=True)
    save_setup_config(serve_path, indexed_labels.keys(), args)
    save_index_to_name(serve_path, indexed_labels)
    save_sample(serve_path)
    save_handler(serve_path)


def create_mar(
    args: Namespace,
    model_dir: Optional[Path] = None,
    model_name: Optional[str] = None,
    version: Optional[float] = 1.0,
    serialized_file: Optional[str] = None,
    handler: Optional[str] = "transformer_handler_pq.py",
):
    """build a torch-model-archive file using
    https://github.com/pytorch/serve/tree/master/model-archiver

    Args:
        args (Namespace): the project argparse namespace.
        model_dir (Path, optional): the directory containing
        mar inputs, and the output directory. Defaults to None.
        model_name (str, optional): the name of the model to be served.
        Defaults to None.
        version (float, optional): the model version. Defaults to 1.0.
        serialized_file (str, optional): the model output of save_pretrained().
        Defaults to None.
        handler (str, optional): the serving handler file.
        Defaults to "transformer_handler_pq.py".
    """
    if model_dir is None:
        model_dir = Path(io.id_str("", args)).parent.joinpath("serve")
    export_dir = model_dir.joinpath("mar")
    export_dir.mkdir(parents=True, exist_ok=True)
    xfiles = ",./".join(shlex.quote(x) for x in EXTRA_FILES)
    if model_name is None:
        model_name = args.experiment.replace("-", "_")
    if serialized_file is None:
        serialized_file = "pytorch_model.bin"
    if mar_files(model_dir, serialized_file, handler):
        # --model-file=./model.py
        cmd = f"""torch-model-archiver
            --model-name={model_name}
            --version={version}
            --serialized-file=./{serialized_file}
            --handler=./{handler}
            --extra-files "./{xfiles}"
            --export-path={export_dir}
        """
        sp = subprocess.Popen(shlex.split(cmd), cwd=model_dir)
        sp.communicate()
        print(f"torch archive {model_name}.mar created")



import pytorch_quik as pq
from collections import OrderedDict
import torch
import numpy as np
import pandas as pd
import pytest


def test_dir_classes(senti_classes):
    raw_classes = list(senti_classes.values())
    dir_classes = pq.metrics.direct_dict(raw_classes)
    assert isinstance(dir_classes, OrderedDict)
    assert dir_classes == senti_classes


def test_inv_classes(senti_classes, inv_senti_classes):
    inv_classes = pq.metrics.inverse_dict(senti_classes)
    assert isinstance(inv_classes, OrderedDict)
    assert inv_classes == inv_senti_classes


def test_choose_a_class():
    cc = torch.rand(100, 100)
    cc = pq.metrics.choose_a_class(cc)
    assert cc.shape == torch.Size([100])


def test_numpize_array(args):
    ln = 100
    wd = 3
    t = torch.rand(ln, wd)
    p = pd.DataFrame(np.random.randint(0, ln, size=(ln, wd)))
    s = pd.Series(np.random.randint(0, ln, size=ln))
    objs = [t, p, s]
    if bool(args.gpus) and not args.has_gpu:
        with pytest.raises(ImportError):
            import cudf
    elif args.gpus == 1:
        import cudf
        c = cudf.DataFrame.from_pandas(p)
        objs.append(c)
    assert all(
        [isinstance(pq.metrics.numpize_array(x), np.ndarray) for x in objs]
    )


def test_show_confusion_matrix(
    final_cmdf, sample_preds, sample_labels, senti_classes
):
    cm = pq.metrics.show_confusion_matrix(
        sample_preds, sample_labels, senti_classes)
    assert cm.equals(final_cmdf)


@pytest.mark.mpl_image_compare
def test_matplotlib_confusion_matrix(final_cmdf, senti_classes):
    npdf = final_cmdf.to_numpy()
    cm = pq.metrics.matplotlib_confusion_matrix(npdf, senti_classes)
    return cm



import pytorch_quik as pq
from pathlib import Path


def test_id_str(args):
    """print an id str"""
    filename = "train_tensor_20210101.pt"
    filename = Path.cwd().joinpath("data", args.bert_type, filename)
    # assert(pq.utils.id_str("train", args) == output)
    output = pq.io.id_str("train", args)
    print(output)
    assert filename == output



import torch
from argparse import Namespace
from typing import Optional, Union
from pathlib import Path
from datetime import date
import json
import numpy as np


def id_str(
    ftype: str,
    args: Namespace,
    epoch: Optional[str] = None,
    gpu: Optional[str] = None,
    suffix: Optional[str] = ".pt",
) -> str:
    """Method to determine an appropriate filename for tensors, models, and
    state_dicts

    Args:
        ftype (str): file type (state_dict, train, valid, test, preds, model)
        args (Namespace): the current set of arguments. requires date,
            optional bert_type, source
        epoch (Optional[str], optional): the current epoch to be saving.
            Defaults to None.
        gpu (Optional[str], optional): the current gpu (for preds). Defaults
            to None.
        suffix (Optional[str], optional): file extension. Defaults to ".pt".

    Returns:
        str: full filename for output object
    """
    lbls = "".join(map(str, getattr(args, "labels", "")))
    if ftype == "state_dict":
        epoch = "e" + str(epoch)
    elif ftype == "preds":
        suffix = ".csv"
    elif ftype == "confusion":
        suffix = ".png"
    elif ftype == "test_array":
        suffix = ".npy"
    elif ftype != "model":
        ftype = ftype + "_tensor"
    data_date = getattr(args, "data_date", date.today().strftime("%Y%m%d"))
    id_list = [ftype, lbls, epoch, str(data_date), gpu]
    id_str = "_".join(filter(None, id_list))
    path_list = [
        "data",
        getattr(args, "bert_type", None),
        getattr(args, "source", None),
        id_str,
    ]
    path_list = filter(None, path_list)
    filename = Path.cwd().joinpath(*path_list).with_suffix(suffix)
    filename.parent.mkdir(parents=True, exist_ok=True)
    return filename


def load_torch_object(
    torch_type: str,
    args: Namespace,
    epoch: Optional[int] = None,
    location: Optional[bool] = False,
) -> Union[str, torch.Tensor]:
    """Load a PyTorch object (tensor, model, state_dict)

    Args:
        ttype (str): The type of torch object (train, valid, test, state_dict)
        args (Namespace): The argparse namespace
        epoch (int, optional): Current epoch (for state_dict)
        location (bool, optional): If the object isn't wanted, but
        only the location (filename). Defaults to False.

    Returns:
        Union[str, torch.Tensor]: [description]
    """

    print("loading " + torch_type + " " + str(getattr(args, "device", "cpu")))
    filename = id_str(torch_type, args, epoch)
    if location:
        return filename
    pt = torch.load(filename)
    if torch_type == "state_dict":
        # DDP will leave module artifacts to be removed
        pt = {k.replace("module.", ""): v for k, v in pt.items()}
    return pt


def load_tensor(
    ttype: str, args: Namespace, loc_only: Optional[bool] = False
) -> Union[str, torch.Tensor]:
    """Load a PyTorch tensor

    Args:
        ttype (str): The type of tensor (train, valid, test)
        args (Namespace): The argparse namespace
        loc_only (bool, optional): If the tensor isn't wanted, but
        only the location (filename). Defaults to False.

    Returns:
        Union[str, torch.Tensor]: [description]
    """
    print("loading " + ttype + " " + str(args.device))
    file_id = id_str(ttype, args)
    if loc_only:
        return file_id
    else:
        return torch.load(file_id)


# def load_state_dict():
#     sd_id = pq.utils.id_str("state_dict", args, epoch)
#     state_dict = torch.load(sd_id, map_location=self.gpus.device)
#     # DDP will leave module artifacts to be removed
#     ddp_state_dict = {
#         k.replace("module.", ""): v for k, v in state_dict.items()
#     }
#     self.model.load_state_dict(ddp_state_dict)


def save_state_dict(model, args: Namespace, epoch: int):
    sd_id = id_str("state_dict", args, epoch)
    print("saving epoch " + str(epoch) + " state")
    if hasattr(model, "module"):
        sd = model.module.state_dict()
    else:
        sd = model.state_dict()
    torch.save(sd, sd_id)
    return sd_id


def save_test_array(Xte: np.array, yte: np.array, args: Namespace):
    filename = id_str("test_array", args)
    with open(filename, 'wb') as f:
        np.save(f, Xte)
        np.save(f, yte)


def load_test_array(args):
    filename = id_str("test_array", args)
    with open(filename, 'rb') as f:
        Xte = np.load(f, allow_pickle=True)
        yte = np.load(f, allow_pickle=True)
    return Xte, yte


def json_write(serve_path, filename, data):
    file = Path.joinpath(serve_path, filename)
    with open(file, 'w') as outfile:
        json.dump(data, outfile)



import time
from typing import Union, Dict, Tuple, OrderedDict


def sec_str(st: float) -> str:
    """Simple util that tells the amount of time for a codeset (in seconds)

    Args:
        st (time): The start time of the codeset

    Returns:
        str: time in seconds
    """
    return str(round(time.time() - st, 2)) + " seconds"


def row_str(dflen: int) -> str:
    """String wrapper for the million of rows in a dataframe

    Args:
        dflen (int): the length of a dataframe

    Returns:
        str: rows in millions
    """
    return str(round(dflen / 1000000, 1)) + "M rows"


def indexed_dict(classes: Tuple[str]) -> OrderedDict[int, str]:
    """Create an ordered dict of classes with indices as keys

    Args:
        classes (Tuple[str]): A list of classes

    Returns:
        OrderedDict[int, str]: A final ordered dict
    """
    class_keys = range(len(classes))
    return OrderedDict(zip(class_keys, classes))


def inverse_dict(
    dict_direct: Union[Dict, OrderedDict]
) -> OrderedDict[str, int]:
    """invert a dictionary

    Args:
        dict_direct (Union[Dict, OrderedDict]): the original
            dictionary to be inverted
    Returns:
        OrderedDict[str, int]: The inverted ordered dictionary
    """
    dict_inverse = {v: k for k, v in dict_direct.items()}
    return OrderedDict(dict_inverse)



import multiprocessing as mp
from urllib3.util.retry import Retry
from requests.adapters import HTTPAdapter
from requests import Session
from typing import List, OrderedDict
from multiprocessing.connection import Connection
import pandas as pd
import numpy as np
from math import ceil
import json
from pandas.io.json import json_normalize

URL = "http://deepshadow.gsslab.rdu2.redhat.com:8080/predictions/my_tc"
JSON_HEADER = {"Content-Type": "application/json"}


def requests_session() -> Session:
    retry_strategy = Retry(
        total=3,
        backoff_factor=1,
        status_forcelist=[507],
        method_whitelist=["POST"],
    )
    adapter = HTTPAdapter(max_retries=retry_strategy)
    sesh = Session()
    sesh.mount("http://", adapter)
    return sesh


def request_post(data: str, sesh: Session, conn: Connection, num: int):
    r = sesh.post(URL, data=bytes(data, "utf-8"), headers=JSON_HEADER)
    print(f"{num} status_code: {r.status_code}")
    conn.send(r)


def split_and_format(arr: np.array, length: int) -> List[List[str]]:
    splits = ceil(len(arr) / length)
    arr_list = np.array_split(arr.flatten(), splits)
    data_list = [
        f'{{"instances":{json.dumps([{"data": text} for text in arr])}}}'
        for arr in arr_list
    ]
    return data_list


def batch_inference(
    responses: np.array, indexed_labels: OrderedDict, batch_size: int
) -> pd.DataFrame:
    data_list = split_and_format(responses, batch_size)
    processes = []
    r_list = []
    sesh = requests_session()
    for num, batch in enumerate(data_list):
        output_pipe, input_pipe = mp.Pipe(duplex=False)
        proc = mp.Process(
            target=request_post, args=(batch, sesh, input_pipe, num)
        )
        processes.append(proc)
        r_list.append(output_pipe)
        proc.start()
    [proc.join() for proc in processes]
    r_list = [
        json_normalize(r.recv().json()["predictions"], sep="_") for r in r_list
    ]
    return pd.concat(r_list, ignore_index=True)



import numpy as np
from torch import Tensor
from transformers import (
    BertTokenizer,
    RobertaTokenizer,
    BertForSequenceClassification,
    RobertaForSequenceClassification,
    BatchEncoding,
    PreTrainedTokenizer,
    PreTrainedModel
)
from typing import Optional, List, Union, Dict
import shutil
from pathlib import Path
import json
from . import io, ddp
from argparse import Namespace
from torch.nn.parallel import DistributedDataParallel as DDP

Tokenizers = Union[BertTokenizer, RobertaTokenizer]
Models = Union[BertForSequenceClassification, RobertaForSequenceClassification]

BERT_MODELS = {
    "bert": {
        "model-type": "bert-base-uncased",
        "tokenizer": BertTokenizer,
        "model": BertForSequenceClassification,
    },
    "roberta": {
        "model-type": "roberta-base",
        "tokenizer": RobertaTokenizer,
        "model": RobertaForSequenceClassification,
    },
}

BATCH_ENCODE_KWARGS = {
    "add_special_tokens": True,
    "return_attention_mask": True,
    "pad_to_max_length": True,
    "max_length": 256,  # max is 512
    "truncation": True,
    "return_tensors": "pt",
}


def get_bert_info(
    labels: Tensor, bert_type: Optional[str] = "bert"
) -> Dict[str, Union[str, bool]]:

    """Get a BERT or RoBERTa tokenizer from transformers to create
    encodings.
    Args:
        labels (Tensor): A tensor of the labels in the dataset. This will
        affect the size of the network final layer
        bert_type (str, optional): The type of the model to be used ("bert" or
        "roberta"). Defaults to "bert".

    Returns:
        Dict [str, Any]: A dict that could be provided for torch serve
    """
    bert_dict = BERT_MODELS[bert_type]
    num_lbls = len(np.unique(labels))
    serve_config = {
        "model_name": bert_dict["model-type"],
        "mode": "sequence_classification",
        "do_lower_case": True,
        "num_labels": str(num_lbls),
        "save_mode": "pretrained",
        "max_length": BATCH_ENCODE_KWARGS["max_length"],
        "captum_explanation": False,
        "embedding_name": bert_type,
    }
    return serve_config


def get_tokenizer(
    bert_type: Optional[str] = "bert", cache_dir: Optional[str] = None
) -> Tokenizers:
    """Get a BERT or RoBERTa tokenizer from transformers to create
    encodings.

    Args:
        bert_type (str, optional): The type of tokenizer ("bert" or "roberta").
        Defaults to "bert".

    Returns:
        Tokenizers: The tokenizer to be used by get_encodings.
    """
    bert_dict = BERT_MODELS[bert_type]
    tokenizer = bert_dict["tokenizer"].from_pretrained(
        pretrained_model_name_or_path=bert_dict["model-type"],
        do_lower_case=True,
        cache_dir=cache_dir,
    )
    return tokenizer


def save_tokenizer(tokenizer, data_path):
    serve_path = data_path.joinpath("serve")
    tokenizer.save_pretrained(serve_path)
    configpath = Path(serve_path).joinpath("tokenizer_config.json")
    newtokenizer = "./tokenizer.json"
    with open(configpath) as f:
        tconfig = json.load(f)
    srcpath = Path(tconfig["tokenizer_file"])
    dstpath = serve_path.joinpath(newtokenizer)
    shutil.copyfile(srcpath, dstpath)
    tconfig["tokenizer_file"] = newtokenizer
    io.json_write(serve_path, configpath, tconfig)


def save_bert_model(
    model: PreTrainedModel,
    args: Namespace,
    best_epoch: int,
    serve_path: Optional[Path] = None
):
    if isinstance(model, DDP):
        model = model.module
    if serve_path is None:
        serve_path = io.id_str("", args).parent
    state_dict = io.load_torch_object("state_dict", args, best_epoch)
    if 'module.classifier.bias' in state_dict:
        print("yes module")
        state_dict = ddp.consume_prefix_in_state_dict_if_present(
            state_dict,
            "module."
        )
    model.save_pretrained(save_directory=serve_path, state_dict=state_dict)


def get_encodings(
    array_list: List[np.ndarray],
    idx: int,
    bert_type: Optional[str] = "bert",
    tokenizer: Optional[PreTrainedTokenizer] = None,
) -> List[BatchEncoding]:
    """Get BERT or RoBERTa encodings for a list of np.ndarrays for training,
    testing (and validation).

    Args:
        array_list (List[np.ndarray]): The list of training, test, (and
        validation) arrays.
        idx (int): The column index of the text in the original pd.DataFrame
        bert_type (str, optional): The type of bert model to be used. Defaults
        to "bert".
        tokenizer (PreTrainedTokenizer, optional): The tokenizer to use

    Returns:
        List[BatchEncoding]: A list of training, test (and validation)
        encodings of text.
    """
    if tokenizer is None:
        tokenizer = get_tokenizer(bert_type)
    encode_list = [
        tokenizer(list(array[:, idx]), **BATCH_ENCODE_KWARGS)
        for array in array_list
    ]
    return encode_list, tokenizer


def get_pretrained_model(
    labels: Tensor, bert_type: Optional[str] = "bert"
) -> Models:
    """Get a pretrained model from transformers.

    Args:
        labels (Tensor): A tensor of the labels in the dataset. This will
        affect the size of the network final layer
        bert_type (str, optional): The type of the model to be used ("bert" or
        "roberta"). Defaults to "bert".

    Returns:
        Models: The pretrained model to be used in training.
    """
    num_lbls = len(np.unique(labels))
    bert_dict = BERT_MODELS[bert_type]
    model = bert_dict["model"].from_pretrained(
        bert_dict["model-type"],
        num_labels=num_lbls,
        output_attentions=False,
        output_hidden_states=False,
    )
    return model



import pytorch_quik as pq
from time import time, sleep
import re


def test_sec_str():
    """print a sec_str"""
    start_time = time()
    sleep(2)
    msg = pq.utils.sec_str(start_time)
    assert re.match(r"^2.0\d? seconds$", msg)


def test_row_str(sample_data):
    """print a row_str"""
    assert pq.utils.row_str(sample_data.shape[0]) == "0.0M rows"



from sklearn.metrics import f1_score, confusion_matrix, classification_report
import matplotlib.pyplot as plt
import matplotlib.colors as mclr

# from collections import OrderedDict
from typing import Union, Optional, List, OrderedDict
import time
from datetime import timedelta
import numpy as np
import pandas as pd
from torch import Tensor
import torch
from pytorch_quik import utils
from pathlib import Path

try:
    import cudf
except ImportError:
    import dask_quik.dummy as cudf

ArrType = Union[
    np.ndarray, pd.Series, pd.DataFrame, cudf.DataFrame, torch.Tensor
]
LABELS = ["Actual", "Predicted"]


def choose_a_class(preds_array: ArrType) -> ArrType:
    """Given an np.array with probabilities for multiple classes,
    provide the argmax class back as a np.array

    Args:
        preds_array (ArrType): 2d array with class probabilities

    Returns:
        np.array: 1d np.array with selected class
    """
    return np.argmax(preds_array, axis=1).flatten()


def numpize_array(arr: ArrType) -> np.array:
    """Given some type of array, try to convert it to a
    numpy array, and of one dimension

    Args:
        arr (ArrType): 1d or 2d array of some type

    Returns:
        np.array: 1d numpy array
    """
    if isinstance(arr, cudf.DataFrame):
        arr = arr.as_matrix()
    elif isinstance(arr, (pd.DataFrame, pd.Series)):
        arr = arr.to_numpy()
    elif isinstance(arr, torch.Tensor):
        if arr.device.type == 'cuda':
            arr = arr.cpu()
        arr = arr.numpy()
    if len(arr.shape) == 2:
        if arr.shape[1] == 1:
            arr = arr.squeeze()
        else:
            arr = choose_a_class(arr)
    return arr


def accuracy_per_class(preds_array: ArrType, true_array: ArrType, label_dict):
    """Accuracy per class function taken from Ali Anastassiou's
    BERT course
    https://www.coursera.org/projects/sentiment-analysis-bert

    Args:
        preds_array (ArrType): [description]
        true_array (ArrType): [description]
        label_dict ([type]): [description]
    """
    # preds_array = choose_a_class(preds_array)
    [parr, tarr] = [numpize_array(x) for x in [preds_array, true_array]]
    for nclass in np.unique(tarr):
        y_preds = parr[tarr == nclass]
        y_true = tarr[tarr == nclass]
        print(f"Class: {label_dict[nclass]}")
        print(f"Accuracy: {len(y_preds[y_preds==nclass])}/{len(y_true)}\n")


def f1_score_func(preds_array: ArrType, true_array: ArrType) -> np.float64:
    """F1 score function taken from Ali Anastassiou's BERT course
    https://www.coursera.org/projects/sentiment-analysis-bert


    Args:
        preds_array (ArrType): Prediction array
        true_array (ArrType): True labels array

    Returns:
        np.float64: F1 score
    """
    [parr, tarr] = [numpize_array(x) for x in [preds_array, true_array]]
    f1 = f1_score(tarr, parr, average="weighted")
    print(f"weighted f1 score {f1:.3f}")


def matplotlib_gist_heat():
    cmap = plt.get_cmap("gist_heat")
    new_cmap = mclr.LinearSegmentedColormap.from_list(
        "trunc({n},{a:.2f},{b:.2f})".format(n=cmap.name, a=0.01, b=0.9),
        cmap(np.linspace(0.01, 0.9, 100)),
    )
    return new_cmap


def matplotlib_confusion_matrix(
    conf_matrix: np.ndarray,
    dir_classes: OrderedDict[int, str],
    save_loc: Optional[Path] = Path.cwd(),
    show_plot: Optional[bool] = False,
):
    """Turn a confusion matrix into a matplotlib plot.

    Args:
        conf_matrix (np.ndarray): a 2d array of the confusion matrix
        dirclasses (OrderedDict[int, str]): A list of chart labels for each
            classification type
    """
    heat_cmap = matplotlib_gist_heat()
    ticks = list(dir_classes.keys())
    classes = list(dir_classes.values())
    lw = len(classes)
    fig, ax = plt.subplots(figsize=(lw, lw))
    plt.imshow(conf_matrix, cmap=heat_cmap)
    plt.xticks(ticks=ticks, labels=classes)
    plt.yticks(ticks=ticks, labels=classes, rotation=90)
    _ = [
        ax.text(j, i, conf_matrix[i, j], ha="center", va="center", color="w")
        for i in range(lw)
        for j in range(lw)
    ]
    plt.ylabel(LABELS[0])
    plt.xlabel(LABELS[1])
    if show_plot:
        plt.show()
    plt.savefig(save_loc, bbox_inches="tight", dpi=150)


def pandas_confusion_matrix(
    conf_matrix: np.ndarray,
    dir_classes: OrderedDict[int, str],
):
    """Turn a confusion matrix into a matplotlib plot.

    Args:
        conf_matrix (np.ndarray): a 2d array of the confusion matrix
        dirclasses (OrderedDict[int, str]): A list of chart labels for each
            classification type
    """
    classes = dir_classes.values()
    labels = [pd.MultiIndex.from_product([[lbl], classes]) for lbl in LABELS]
    pdcm = pd.DataFrame(conf_mat, *labels)
    return pdcm


def build_confusion_matrix(
    preds_array: ArrType,
    true_array: ArrType,
    dir_classes: OrderedDict[int, str],
    save_loc: Optional[Path] = Path.cwd(),
    show_plot: Optional[bool] = False,
    in_pandas: Optional[bool] = False,
) -> np.ndarray:
    """Takes either a 2d classification array or a 1d array and true
    values, and creates a confusion matrix, both with seaborn and pandas.
    Taken from https://stackoverflow.com/questions/65390832/
    set-meta-name-for-rows-and-columns-in-pandas-dataframe

    Args:
        preds_array (ArrType): A 2d np.array to be argmaxed, a
            pd.Series, or a 1d np.array
        true_array (ArrType): A pd.Series or 1d np.array
        dir_classes: OrderedDict[int, str]: A dict of index keys and
            labels for each classification type
        plot (bool, optional): If this is in a notebook, then you may
        wish to create a seaborn plot. Defaults to False.
    """
    [parr, tarr] = [numpize_array(x) for x in [preds_array, true_array]]
    conf_mat = confusion_matrix(tarr, parr)
    matplotlib_confusion_matrix(conf_mat, dir_classes, save_loc, show_plot)
    if in_pandas:
        return pandas_confusion_matrix(conf_mat)
    return conf_mat


def build_class_dict(
    preds_array: ArrType,
    true_array: ArrType,
    dir_classes: OrderedDict[int, str],
    print_report: Optional[bool] = False,
):
    """Turn a confusion matrix into a classifcation report.

    Args:
        preds_array (ArrType): A 2d np.array to be argmaxed, a
            pd.Series, or a 1d np.array
        true_array (ArrType): A pd.Series or 1d np.array
        dir_classes: OrderedDict[int, str]: A dict of index keys and
            labels for each classification type
    """
    [parr, tarr] = [numpize_array(x) for x in [preds_array, true_array]]
    tn = utils.inverse_dict(dir_classes)
    if print_report:
        print(classification_report(tarr, parr, target_names=tn))
    class_rpt = classification_report(
        tarr, parr, target_names=tn, output_dict=True
    )
    wa = class_rpt["weighted avg"].copy()
    poprows = ["accuracy", "macro avg", "weighted avg"]
    [class_rpt.pop(key) for key in poprows]
    popcols = ["precision", "recall", "support"]
    [v.pop(col) for col in popcols for v in class_rpt.values()]
    class_rpt["weighted_avg"] = wa
    class_rpt = pd.json_normalize(class_rpt, sep="_").to_dict(
        orient="records"
    )[0]
    return class_rpt


class SmoothenLoss:
    """Taken from fast.ai, Create a smooth moving average with a beta
    value (https://github.com/fastai/fastai1/blob/master/fastai/callback.py)"""

    def __init__(self, beta: float):
        """Create the smoothen value instance for loss metrics

        Args:
            beta (float): The smoothing beta
        """
        self.beta = beta
        self.n = 0
        self.mov_avg = 0

    def add_value(self, val: float) -> None:
        """Add value to calculate updated smoothed value.

        Args:
            val (float): The value to be added
        """

        self.n += 1
        self.mov_avg = self.beta * self.mov_avg + (1 - self.beta) * val
        self.smooth = self.mov_avg / (1 - self.beta ** self.n)


class LossMetrics:
    """Don's method to store dict without callbacks, adapted from
    https://github.com/fastai/fastai/blob/master/fastai/metrics.py"""

    def __init__(self, beta: Optional[float] = 0.99):
        """Create the loss metrics structure

        Args:
            beta (float): A beta for smoothing the average. Default is 0.99
        """
        self.beta = beta
        self.smoothener = SmoothenLoss(self.beta)
        self.metric_dict = {"epoch": 0, "train_loss": 0.0, "valid_loss": 0.0}
        self.results = pd.DataFrame()
        self.start_time = time.time()

    def add_loss(self, loss: Tensor):
        """Handle gradient calculation on loss.

        Args:
            loss (Tensor): The loss tensor from criterion
        """
        loss = loss.float().detach().cpu().item()
        self.smoothener.add_value(loss)
        sss = self.smoothener.smooth
        self.metric_dict["train_loss"] = sss

    def add_vloss(self, vlosses: List[Tensor], nums: List[int]):
        """Handle validation loss calculation

        Args:
            vlosses (List[Tensor]): A set of validation tensors
            nums (List[int]): The length of the groups
        """
        nums = np.array(nums, dtype=np.float32)
        fl = (
            torch.stack(vlosses).data.cpu().numpy() * nums
        ).sum() / nums.sum()
        self.metric_dict["valid_loss"] = fl

    def calc_time(self):
        """Add time to the metrics DataFrame"""
        elapsed = time.time() - self.start_time
        self.duration = (
            str(timedelta(seconds=elapsed)).split(".")[0].partition(":")[2]
        )
        self.metric_dict["time"] = self.duration

    def write(self):
        """write out the metrics dataframe when completed"""
        self.calc_time()
        self.results = pd.concat(
            [
                self.results,
                pd.DataFrame(
                    self.metric_dict, index=[self.metric_dict["epoch"]]
                ),
            ]
        )
        self.metric_dict["epoch"] += 1



from pathlib import Path
import pytest
import pandas as pd
import numpy as np
import torch
import json
from argparse import Namespace
from os import system
import warnings
from collections import OrderedDict

TESTDIR = Path(__file__).parent
SAMPLE = TESTDIR.joinpath("sample_data.json")
FINAL = TESTDIR.joinpath("final_data.json")


# def pytest_generate_tests(metafunc):
#     metafunc.parametrize("gpus", [0, 1])
@pytest.fixture(params=[0, 1])
def gpus(request):
    return request.param


@pytest.fixture
def args(gpus):
    """sample args namespace"""
    args = Namespace()
    args.gpus = gpus
    args.data_date = 20210101
    args.epochs = 2
    args.has_gpu = system("nvidia-smi -L") == 0
    if not args.has_gpu:
        warnings.warn(
            "GPU not found, setting has_gpu to False. \
            Some tests will be skipped"
        )
        args.device = torch.device('cpu')
    else:
        args.device = torch.device('cuda', 0)
    args.nr = 0
    args.nodes = 1
    args.bs = 12
    args.num_workers = 0
    args.learning_rate = 5e-5
    args.lr = 5e-5
    args.eps = 1e-4
    args.betas = (0.90, 0.99)
    args.weight_decay = 0
    args.mixed_precision = True
    args.find_unused_parameters = False
    args.bert_type = "roberta"
    return args


@pytest.fixture(scope='session')
def senti_classes():
    """sentiment classes"""
    dir_classes = OrderedDict([(0, 'Negative'), (1, 'Neutral'), (2, 'Positive')])
    return dir_classes


@pytest.fixture(scope='session')
def inv_senti_classes():
    """inverse sentiment classes"""
    inv_classes = OrderedDict([('Negative', 0), ('Neutral', 1), ('Positive', 2)])
    return inv_classes


@pytest.fixture(scope="session")
def sample_tensor():
    """sample tensor dataset"""
    torch.manual_seed(0)
    return torch.rand(200, 64)


@pytest.fixture(scope="session")
def sample_preds():
    """labels tensor"""
    torch.manual_seed(0)
    p = torch.rand(200, 3)
    # https://stackoverflow.com/questions/59090533/how-do-i-add-some-gaussian
    # -noise-to-a-tensor-in-pytorch
    noise = torch.zeros(200, 3, dtype=torch.float64)
    noise = noise + (0.1**0.9)*torch.randn(200, 3)
    return p + noise


@pytest.fixture(scope="session")
def sample_labels():
    """labels tensor"""
    torch.manual_seed(0)
    labels = torch.rand(200, 3)
    labels = np.argmax(labels, axis=1).flatten()
    # torch.manual_seed(0)
    # return (torch.rand(200) < 0.5).int()
    return labels


@pytest.fixture(scope="session")
def sample_data():
    """sample user/item dataset"""
    with open(SAMPLE) as f:
        df = pd.DataFrame(json.load(f))
    return df


@pytest.fixture(scope="session")
def final_data():
    """final user/item dataset"""
    with open(FINAL) as f:
        df = pd.DataFrame(json.load(f))
    df.index.names = ["ui_index"]
    return df


@pytest.fixture(scope="session")
def final_cmdf(senti_classes):
    """final confusion matrix data frame"""
    arr = np.array([[63,  8,  7], [6, 48,  4], [2,  8, 54]])
    classes = senti_classes.values()
    aidx = pd.MultiIndex.from_product([['Actual'], classes])
    pidx = pd.MultiIndex.from_product([['Predicted'], classes])
    df = pd.DataFrame(arr, aidx, pidx)
    return df


@pytest.fixture(scope="session")
def ttypes():
    """tensor data types for transformation"""
    ttypes = {
        0: torch.int16,
        1: torch.int16
    }
    return ttypes



from argparse import ArgumentParser
import dask_quik as dq


def beta_type(strings: str) -> tuple:
    """An argparse type function that takes a string and returns
    a two part tuple for a low and high beta for an optimizer

    Args:
        strings (str): a comma separated string of numbers

    Returns:
        tuple: a tuple of floats
    """
    mapped_beta = map(float, strings.split(","))
    return tuple(mapped_beta)


def add_learn_args(parser: ArgumentParser, kwargs={}) -> ArgumentParser:
    """Add learn args will add arguments that are common to PyTorch, and
    necessary for a pytorch-quik traveler. These can be set by command
    line, or can be defaulted in a script, or the defaults here can be
    used.

    Args:
        parser (ArgumentParser): This is the parser from the PyTorch project.
        kwargs (dict, optional): Sometimes the individual project will want
        to have default learning args. For instance, learning rate could be
        set at the command line, if not by the PyTorch project, or if not it
        will be set here. If they are all set here, then kwargs defaults to {}.

    Returns:
        ArgumentParser: The same ArgumentParser but with loaded learning args.
    """
    parser.add_argument(
        "-e",
        "--epochs",
        default=kwargs.get("epochs", 5),
        type=int,
        metavar="N",
        help="number of total epochs to run (2, 3, 5)",
    )    
    parser.add_argument(
        "-nr",
        "--nr",
        default=0,
        type=int,
        help="ranking within the nodes",
    )
    parser.add_argument(
        "-n",
        "--nodes",
        default=1,
        type=int,
        metavar="N",
        help="number of data loading workers (default: 4)",
    )
    parser.add_argument(
        "-g",
        "--gpus",
        default=dq.utils.gpus(),
        type=int,
        help="number of gpus per node",
    )
    parser.add_argument(
        "-amp",
        "--mixed_precision",
        action="store_true",
        default=True,
    )
    parser.add_argument(
        "-bs",
        "--bs",
        default=kwargs.get("bs", 50000),
        type=int,
        help="batch size (16, 32, 150000)",
    )
    parser.add_argument(
        "-lr",
        "--learning_rate",
        dest="lr",
        default=kwargs.get("learning_rate", 1e-5),
        type=float,
        help="learning rate for an optimizer (1e-5, 2e-6)",
    )
    parser.add_argument(
        "-wd",
        "--weight_decay",
        default=kwargs.get("weight_decay", 0),
        type=float,
        help="weight decay for an optimizer",
    )
    parser.add_argument(
        "--betas",
        type=beta_type,
        default=kwargs.get("betas", (0.9, 0.99)),
        help="beta values for an optimizer. Enter them as \
            one string, such as '0.90, 0.99'.",
    )
    parser.add_argument(
        "-eps",
        default=kwargs.get("eps", 1e-08),
        type=float,
        help="weight decay for an optimizer",
    )
    parser.add_argument(
        "-nw",
        "--num_workers",
        # default=torch.get_num_threads(),
        default=kwargs.get("num_workers", 0),
        type=int,
        help="number of workers",
    )
    parser.add_argument(
        "--find_unused_parameters",
        dest="find_unused_parameters",
        action="store_true",
    )
    return parser


def add_mlflow_args(parser: ArgumentParser, kwargs={}) -> ArgumentParser:
    """Add mlflow args will add arguments that are common to mlflow, and
    necessary for a pytorch-quik traveler. These can be set by command
    line, or can be defaulted in a script, or the defaults here can be
    used.

    Args:
        parser (ArgumentParser): This is the parser from the PyTorch project.
        kwargs (dict, optional): Sometimes the individual project will want
        to have default learning args. For instance, learning rate could be
        set at the command line, if not by the PyTorch project, or if not it
        will be set here. If they are all set here, then kwargs defaults to {}.

    Returns:
        ArgumentParser: The same ArgumentParser but with loaded mlflow args.
    """
    parser.add_argument(
        "--use_mlflow",
        dest="use_mlflow",
        action="store_true",
    )
    parser.add_argument(
        "-exp",
        "--experiment",
        default=kwargs.get("experiment", "Default"),
        type=str,
        help="The name of the mlflow experiment (default is Default)",
    )
    parser.add_argument(
        "--tracking_uri",
        default=kwargs.get("tracking_uri", "http://localhost:5000"),
        type=str,
        help="The tracking uri for mlflow (default is localhost, port 5000)",
    )
    parser.add_argument(
        "--endpoint_url",
        default=kwargs.get("tracking_uri", "http://s3.amazonaws.com"),
        help="The S3 endpoint where MLFlow will send your model artifacts",
    )
    return parser



import pytorch_quik.api
import pytorch_quik.arg
import pytorch_quik.ddp
import pytorch_quik.io
import pytorch_quik.metrics
import pytorch_quik.serve
import pytorch_quik.tensor
import pytorch_quik.transform
import pytorch_quik.travel
import pytorch_quik.utils

__all__ = [
    "api",
    "arg",
    "io",
    "ddp",
    "metrics",
    "serve",
    "tensor",
    "transform",
    "travel",
    "utils",
]
__version__ = "0.2.0"

try:
    import pytorch_quik.bert

    __all__.append("bert")
except ImportError:
    print("skipping bert functions")



import pytorch_quik as pq
import torch.utils.data as td


def test_make_tds(final_data, args):
    df = final_data.drop('text', axis=1)
    tds = pq.tensor.make_TensorDataset(df, "label")
    assert(isinstance(tds, td.TensorDataset))


def test_transform_tds(final_data, ttypes, args):
    df = final_data.drop('text', axis=1)
    tds = pq.tensor.make_TensorDataset(df, "label")
    tds = pq.tensor.transform_TensorDataset(tds, 1, ttypes)
    assert(isinstance(tds, td.TensorDataset))



import dask_quik as dq
import torch
import torch.multiprocessing as mp
import torch.distributed as dist
from typing import Callable, NamedTuple, Union, Dict, Any
from argparse import Namespace
import os
from tqdm import tqdm


def tq_bar(esteps, epoch=0, total_epochs=0, train=True):
    pbar = tqdm(total=esteps)
    if train:
        pbar.set_description(f"epoch: {epoch +1}/{total_epochs}")
    else:
        pbar.set_description("testing progress")
    return pbar


def traverse(train_fn: Callable, args: Namespace):
    """Determine whether to train using DistributedDataParallel
    or a single training instance

    Args:
        train_fn (Callable): The train function for this particular
        Neural Network
        args (Namespace): The argparse Namespace for this script
    """
    if bool(dq.utils.gpus()):
        ddp_traverse(train_fn, args)
    else:
        train_fn(gpu=None, args=args)


def ddp_traverse(train_fn: Callable, args: Namespace):
    """Using multiprocessing, initiate a DistributedDataParallel
    training process

    Args:
        train_fn (Callable): The train function for this particular
        Neural Network
        args (Namespace): The argparse Namespace for this script
    """
    mp.spawn(train_fn, nprocs=args.gpus, args=(args,))


def setup(gpu: str, args: Union[Namespace, NamedTuple]):
    """Setup of the distrubtion settings.

    Args:
        gpu (str): The current gpu
        args (Union[Namespace, NamedTuple]): The gpu parameters
    """
    os.environ["MASTER_ADDR"] = "localhost"
    os.environ["MASTER_PORT"] = "12355"
    os.environ["NCCL_P2P_LEVEL"] = "0"

    dist.init_process_group(
        backend="nccl",
        init_method="env://",
        world_size=args.world_size,
        rank=args.rank_id,
    )

    # Explicitly setting seed to make sure that models created in two processes
    # start from same random weights and biases.
    torch.manual_seed(42)


def cleanup():
    """Deletion of the DistributedDataParallel process group"""
    dist.destroy_process_group()


# From pytorch 1.9.0
def consume_prefix_in_state_dict_if_present(
    state_dict: Dict[str, Any],
    prefix: str
):
    r"""Strip the prefix in state_dict, if any.
    ..note::
        Given a `state_dict` from a DP/DDP model, a local model can load it by applying
        `consume_prefix_in_state_dict_if_present(state_dict, "module.")` before calling
        :meth:`torch.nn.Module.load_state_dict`.
    Args:
        state_dict (OrderedDict): a state-dict to be loaded to the model.
        prefix (str): prefix.
    """
    keys = sorted(state_dict.keys())
    for key in keys:
        if key.startswith(prefix):
            newkey = key[len(prefix):]
            state_dict[newkey] = state_dict.pop(key)

    # also strip the prefix in metadata if any.
    if hasattr(state_dict, '_metadata'):
        metadata = state_dict._metadata
        for key in list(metadata.keys()):
            # for the metadata dict, the key can be:
            # '': for the DDP module, which we want to remove.
            # 'module': for the actual model.
            # 'module.xx.xx': for the rest.

            if len(key) == 0:
                continue
            newkey = key[len(prefix):]
            metadata[newkey] = metadata.pop(key)
        state_dict._metadata = metadata




import torch
from torch.cuda.amp import GradScaler, autocast
from torch.nn.parallel import DistributedDataParallel as DDP
from torch.nn.utils import clip_grad_norm_
import torch.utils.data.distributed as ddist
import torch.utils.data as td
from torch import nn, optim
from pytorch_quik import arg, ddp, io, metrics
from contextlib import nullcontext
from argparse import ArgumentParser, Namespace
from typing import Any, Callable, Dict, Optional
from dataclasses import dataclass, field, asdict, is_dataclass
import os

try:
    from mlflow.tracking import MlflowClient
except ImportError:
    pass


@dataclass
class World:
    """World related data."""

    device: torch.device = field(init=False)
    node_id: int = 0
    total_nodes: int = 1
    gpu_id: int = None
    total_gpus: int = None
    rank_id: int = field(init=False)
    world_size: int = field(init=False)
    is_ddp: bool = field(init=False)
    is_logger: bool = field(init=False)

    def __post_init__(self):
        if self.gpu_id is None:
            self.device = torch.device("cpu")
            self.rank_id = None
            self.world_size = None
            self.is_ddp = False
        else:
            self.device = torch.device("cuda", self.gpu_id)
            self.rank_id = self.node_id * self.total_gpus + self.gpu_id
            self.world_size = self.total_gpus * self.total_nodes
            self.is_ddp = True
        self.is_logger = not self.is_ddp
        if self.gpu_id == 0:
            self.is_logger = True


@dataclass
class DlKwargs:
    """Data loader keyword arguments."""

    batch_size: int = 24
    shuffle: bool = False
    pin_memory: bool = True
    num_workers: int = 0


@dataclass
class OptKwargs:
    """Optimizer keyword arguments"""

    lr: int
    weight_decay: int
    eps: int
    betas: tuple


class QuikTrek:
    """A class for maintaining the general data for the full trek to
    be shared between travelers.
    """

    def __init__(
        self, args: Optional[Namespace] = None, gpu: Optional[int] = None
    ):
        if args is None:
            parser = arg.add_learn_args(ArgumentParser())
            args = parser.parse_args()
        self.args = args
        self.epochs = args.epochs
        self.create_dataclasses(args, gpu)
        self.trek_prep(args)

    def create_dataclasses(self, args, gpu):
        self.world = World(args.nr, args.nodes, gpu, args.gpus)
        self.dlkwargs = DlKwargs(
            batch_size=args.bs,
            num_workers=args.num_workers,
        )
        self.optkwargs = OptKwargs(
            lr=args.lr,
            weight_decay=args.weight_decay,
            eps=args.eps,
            betas=args.betas,
        )
        self.args.device = self.world.device

    def trek_prep(self, args):
        if args.use_mlflow:
            self.mlflow = QuikMlflow(
                args.experiment,
                args.tracking_uri,
                args.endpoint_url,
                self.world,
                self.dlkwargs,
                self.optkwargs,
            )
        if self.world.device.type == "cuda":
            torch.cuda.empty_cache()
        if self.world.gpu_id is not None:
            torch.cuda.set_device(self.world.device)
            ddp.setup(self.world.gpu_id, self.world)


class QuikTraveler:
    """A class for traversing a model either in training, validation, or
    testing. Is always a singular run - but can be part of a multi-GPU run.
    """

    metrics = metrics.LossMetrics(0.99)

    def __init__(self, trek, type: Optional[str] = None):
        self.type = type
        self.world = trek.world
        self.args = trek.args
        self.trek = trek
        self.find_unused_parameters = trek.args.find_unused_parameters
        self.amp = QuikAmp(trek.args.mixed_precision)

    def set_criterion(
        self,
        criterion_fcn: Callable[..., nn.Module],
        kwargs: Optional[Dict[str, Any]] = {},
    ):
        if callable(criterion_fcn):
            self.criterion = criterion_fcn(**kwargs)
            self.criterion.to(self.world.device)

    def set_optimizer(
        self,
        optimizer_fcn: Callable[..., optim.Optimizer],
        kwargs: Optional[Dict[str, Any]] = {},
    ):
        if hasattr(self.model, "module"):
            self.optimizer = optimizer_fcn(
                self.model.module.parameters(),
                **asdict(self.trek.optkwargs),
                **kwargs,
            )
        else:
            self.optimizer = optimizer_fcn(
                self.model.parameters(),
                **asdict(self.trek.optkwargs),
                **kwargs,
            )

    def set_scheduler(
        self,
        scheduler_fcn: Callable[..., optim.Optimizer],
        kwargs: Optional[Dict[str, Any]] = {},
    ):
        self.scheduler = scheduler_fcn(
            self.optimizer,
            **kwargs,
        )

    def add_model(
        self,
        model: nn.Module,
        state_dict: Optional[Dict[str, torch.Tensor]] = None,
    ):
        self.model = model
        if state_dict is not None:
            self.model.load_state_dict(state_dict)
        self.model.to(self.world.device)
        if self.world.is_ddp:
            self.model = DDP(
                self.model,
                device_ids=[self.world.device],
                output_device=self.world.device,
                find_unused_parameters=self.find_unused_parameters,
            )

    def add_data(self, tensorDataset: td.TensorDataset):
        self.data = QuikData(
            tensorDataset, self.world, self.trek.dlkwargs, self.trek.epochs
        )
        if self.type == "train":
            self.metrics.steps = self.data.steps

    def backward(self, loss: torch.Tensor, clip: Optional[bool] = True):
        if hasattr(self.amp, "scaler"):
            self.amp.backward(self, loss, clip)
        else:
            loss.backward()
            if hasattr(self.amp, "optimizer"):
                self.optimizer.step()

    def add_loss(self, loss, step, epoch):
        self.metrics.add_loss(loss)
        if self.args.use_mlflow:
            loss = self.metrics.metric_dict["train_loss"]
            step = step + (self.metrics.steps * epoch)
            self.trek.mlflow.log_metric("train_loss", loss, step)

    def add_vloss(self, vlosses, nums, epoch):
        self.metrics.add_vloss(vlosses, nums)
        if self.args.use_mlflow:
            vloss = self.metrics.metric_dict["valid_loss"]
            step = self.metrics.steps * (epoch + 1)
            self.trek.mlflow.log_metric("valid_loss", vloss, step)

    def save_state_dict(self, epoch):
        sd_id = io.save_state_dict(self.model, self.args, epoch)
        if self.args.use_mlflow:
            self.trek.mlflow.log_artifact(str(sd_id))

    def record_results(
        self,
        label_names,
        accuracy: Optional[bool] = True,
        f1: Optional[bool] = True,
        confusion: Optional[bool] = True,
    ):
        cm_id = io.id_str("confusion", self.args)
        self.cm = metrics.build_confusion_matrix(
            self.data.predictions,
            self.data.labels,
            label_names,
            cm_id,
        )
        cr = metrics.build_class_dict(
            self.data.predictions,
            self.data.labels,
            label_names,
        )
        self.trek.mlflow.log_artifact(cm_id)
        {
            self.trek.mlflow.log_metric(metric, value, 0)
            for metric, value in cr.items()
        }


class QuikData:
    """A class for providing data to a traveler."""

    def __init__(
        self,
        tensorDataset: td.TensorDataset,
        world: World,
        dlkwargs: Dict[str, Any],
        epochs: int,
    ):
        self.dataset = tensorDataset
        self.labels = self.dataset.tensors[2].cpu().numpy()
        self.dlkwargs = dlkwargs
        self.add_data_loader(world)
        self.steps = len(self.data_loader)
        self.total_steps = self.steps * epochs

    def add_sampler(self, world, sampler_fcn=None, kwargs={}):
        if world.is_ddp:
            self.sampler = ddist.DistributedSampler(
                self.dataset,
                num_replicas=world.world_size,
                rank=world.rank_id,
            )
        elif callable(sampler_fcn):
            self.sampler = sampler_fcn(**kwargs)
        else:
            self.sampler = sampler_fcn

    def add_data_loader(self, world):
        if not hasattr(self, "sampler"):
            self.add_sampler(world)
        self.data_loader = td.DataLoader(
            dataset=self.dataset,
            sampler=self.sampler,
            **asdict(self.dlkwargs),
        )

    def add_results(self, predictions: torch.Tensor, labels: torch.Tensor):
        self.predictions = predictions
        self.labels = labels


class QuikAmp:
    """A class to manage automatic mixed precision. Provides
    a nullcontext to your forward function if it's not being
    used.
    """

    def __init__(self, mixed_precision: bool):
        if mixed_precision:
            self.scaler = GradScaler()
            self.caster = autocast()
        else:
            self.caster = nullcontext()

    def backward(
        self,
        trvlr: QuikTraveler,
        loss: torch.Tensor,
        clip: Optional[bool] = True,
    ):
        self.scaler.scale(loss).backward()
        # https://pytorch.org/docs/stable/notes/amp_examples.html#working-with-unscaled-gradients
        if clip:
            self.scaler.unscale_(trvlr.optimizer)
            # https://discuss.pytorch.org/t/about-torch-nn-utils-clip-grad-norm/13873
            if hasattr(trvlr.model, "module"):
                clip_grad_norm_(trvlr.model.module.parameters(), 1.0)
            else:
                clip_grad_norm_(trvlr.model.parameters(), 1.0)
        self.scaler.step(trvlr.optimizer)
        self.scaler.update()


class QuikMlflow:
    """A class to manage MLflow API calls. You'll have to add your own
    credentials either to ~/.aws/credentials, or store them in environment
    variables like:
        os.environ["AWS_ACCESS_KEY_ID"] = ''
        os.environ["AWS_SECRET_ACCESS_KEY"] = ''
        os.environ["MLFLOW_S3_ENDPOINT_URL"] = ''
    """

    def __init__(
        self,
        experiment,
        tracking_uri,
        endpoint_url,
        world,
        dlkwargs,
        optkwargs,
    ):
        if "MLFLOW_S3_ENDPOINT_URL" not in os.environ:
            os.environ["MLFLOW_S3_ENDPOINT_URL"] = endpoint_url
        self.client = MlflowClient(tracking_uri=tracking_uri)
        exp = self.client.get_experiment_by_name(experiment)
        if exp is None:
            self.expid = self.client.create_experiment(experiment)
        else:
            self.expid = exp.experiment_id
        self.run = self.client.create_run(self.expid)
        self.runid = self.run.info.run_id
        self.log_parameters([world, dlkwargs, optkwargs])

    def log_parameters(self, dclasses):
        for dclass in dclasses:
            if is_dataclass(dclass):
                dclass = asdict(dclass)
            _ = [
                self.client.log_param(self.runid, k, v)
                for k, v in dclass.items()
            ]

    def log_metric(self, key, metric, step):
        self.client.log_metric(self.runid, key, metric, step=step)

    def log_artifact(self, sd_id):
        self.client.log_artifact(self.runid, str(sd_id))

    def end_run(self):
        self.client.set_terminated(self.runid)
